main = ScriptThread("kmcs_mod")

function GUIinit()
	GUI = {}
	GUI.menu = {}
	GUI.buttons = {}
	GUI.navButtons = 	{
					["menuOpen"] = 
					{
						["keys"] = {},
						["menu"] = ""
					}, 
					["menuPrev"] = {["keys"] = {}},
					["navUp"] = {["keys"] = {}},
					["navDown"] = {["keys"] = {}},
					["select"] = {["keys"] = {}},
					["navBack"] = {["keys"] = {}}
				}
	GUI.activeButton = 1
	GUI.activeMenu = ""
	GUI.statusText = {["text"] = "", ["time"] = 0}

	function GUI.createMenu(menuName, prevMenu, title, xpos, ypos, xscale, yscale, buttonSpacing, textScale, font)
		if(GUI.menu[menuName] == nil) then
			GUI.menu[menuName] = {}
			lastOffset = ypos + yscale + buttonSpacing
		end

		GUI.menu[menuName] = {
			["prevMenu"] = prevMenu,
			["title"] = title,
			["xpos"] = xpos,
			["ypos"] = ypos,
			["xscale"] = xscale,
			["yscale"] = yscale,
			["spacing"] = buttonSpacing,
			["lastOffset"] = lastOffset,
			["numButtons"] = 0,
			["textScale"] = textScale,
			["font"] = font
		}
	end

	function GUI.addButton(parentMenu, text, funct, args, toggleable, textScale, font)
		--I tried to make this easy to follow (it probably doesn't help read it, but oh well), but basically,
		--
		if(GUI.buttons[parentMenu] == nil or GUI.buttons[parentMenu]["settings"] == nil) then
			GUI.buttons[parentMenu] = {["settings"] = {}}
		end

		local currButtonNum = #GUI.buttons[parentMenu].settings + 1
		GUI.menu[parentMenu].numButtons = currButtonNum
		--in other words, position the button x.spacing under the previous button

		if(toggleable) then
			toggledOn = false
		else
		    toggledOn = nil
		end

		GUI.buttons[parentMenu].settings[currButtonNum] = {
			["parentMenu"] = parentMenu,
			["text"] = text,
			["funct"] = funct,
			["args"] = args,
			["font"] = font,
			["xpos"] = GUI.menu[parentMenu].xpos,
			["ypos"] = GUI.menu[parentMenu].lastOffset,
			["xscale"] = GUI.menu[parentMenu].xscale,
			["yscale"] = GUI.menu[parentMenu].yscale,
			["textScale"] = textScale,
			["toggleable"] = toggleable,
			["toggledOn"] = toggledOn
		}

		GUI.menu[parentMenu]["lastOffset"] = (GUI.menu[parentMenu].lastOffset + GUI.menu[parentMenu].yscale + GUI.menu[parentMenu].spacing)
	end

	function GUI.setActiveMenu(menuName)
		GUI.activeButton = 1
		GUI.activeMenu = menuName
	end

	function GUI.updateStatusText(text, time)
		GUI.statusText = {["text"] = text, ["time"] = time}
	end

	function GUI.isActiveButtonToggledOn()
		local activeButton = GUI.buttons[GUI.activeMenu].settings[GUI.activeButton]
		if(activeButton.toggleable) then
			return activeButton.toggledOn
		elseif(activeButton.toggleable == false) then
		    return nil
		end
	end

	function GUI.isButtonToggledOn(menu, index)
		local activeButton = GUI.buttons[menu].settings[index]
		if(activeButton.toggleable) then
			return activeButton.toggledOn
		elseif(activeButton.toggleable == false) then
		    return nil
		end
	end

	function GUI.toggleActiveButtonState()
		local activeButton = GUI.buttons[GUI.activeMenu].settings[GUI.activeButton]
		
		if(GUI.isActiveButtonToggledOn() and activeButton.toggleable) then
			activeButton.toggledOn = false
		elseif(GUI.isActiveButtonToggledOn() == false and activeButton.toggleable) then
			activeButton.toggledOn = true
		end
	end

	function GUI.toggleButtonState(menu, index)
		local activeButton = GUI.buttons[menu].settings[index]
		
		if(GUI.isButtonToggledOn(menu, index) and activeButton.toggleable) then
			activeButton.toggledOn = false
		elseif(GUI.isButtonToggledOn(menu, index) == false and activeButton.toggleable) then
			activeButton.toggledOn = true
		end
	end

	--todo
	function GUI.removeButton()
	end

	--todo
	function GUI.removeMenu()
	end

	function GUI.drawStatusText()
		--this is called every frame from GUI.tick()
		--this is pretty much a work around because using a while loop causes the game to freeze/crash
		--and using main:Wait(0) causes the text to flash instead of constant

		if(nextDrawTime == nil or prevText ~= GUI.statusText.text) then
			nextDrawTime = natives.GAMEPLAY.GET_GAME_TIMER() + GUI.statusText.time
		end

		if(nextDrawTime > natives.GAMEPLAY.GET_GAME_TIMER()) then
			prevText = GUI.statusText.text
			natives.UI.SET_TEXT_FONT(0)
			natives.UI.SET_TEXT_SCALE(0.55, 0.55)
			natives.UI.SET_TEXT_COLOUR(255, 255, 255, 255)
			natives.UI.SET_TEXT_WRAP(0.0, 1.0)
			natives.UI.SET_TEXT_CENTRE(true)
			natives.UI.SET_TEXT_DROPSHADOW(0, 0, 0, 0, 0)
			natives.UI.SET_TEXT_EDGE(1, 0, 0, 0, 205)
			natives.UI._SET_TEXT_ENTRY("STRING")
			natives.UI._ADD_TEXT_COMPONENT_STRING(GUI.statusText.text)
			natives.UI._DRAW_TEXT(0.5, 0.5)
		else
			--clear the text on screen when time is up
			GUI.statusText = {["text"] = "", ["time"] = 0}
			nextDrawTime = nil
		end
	end

	function GUI.drawGUI()
		--this is pretty much a work around because using a while loop causes the game to freeze/crash
		--and using main:Wait(0) causes the text to flash instead of constant
		if(GUI.activeMenu ~= "" and GUI.activeMenu ~= nil) then
			local name = GUI.activeMenu

			if(GUI.menu[name] ~= nil) then
				--draw title for one frame
				local xpos = GUI.menu[name].xpos
				local ypos = GUI.menu[name].ypos
				local xscale = GUI.menu[name].xscale
				local yscale = GUI.menu[name].yscale
				
				local text = GUI.menu[name].title
				local textScale = GUI.menu[name].textScale
				local font = GUI.menu[name].font
				local colors = {["r"] = 0, ["g"] = 0, ["b"] = 0, ["o"] = 255}

				addButtonText(text, xpos, ypos, textScale, font)
				natives.GRAPHICS.DRAW_RECT(xpos, ypos, xscale, yscale, colors.r, colors.g, colors.b, colors.o);

				--done drawing the title for that frame, now draw the buttons
				if(GUI.buttons[name] ~= nil) then
					for i, v in pairs(GUI.buttons[name].settings) do
						local buttonSettings = GUI.buttons[name].settings[i]
						local xpos = buttonSettings.xpos
						local ypos = buttonSettings.ypos
						local xscale = buttonSettings.xscale
						local yscale = buttonSettings.yscale
						
						local font = buttonSettings.font
						local text = buttonSettings.text
						local textScale = buttonSettings.textScale
						local colors = {["r"] = 70, ["g"] = 95, ["b"] = 95, ["o"] = 255}

						if(buttonSettings.toggledOn == true) then
							text = buttonSettings.text .. " [ON]"
						elseif(buttonSettings.toggledOn == false) then
							text = buttonSettings.text .. " [OFF]"
						end
						
						if(i == GUI.activeButton) then
							colors = {["r"] = 200, ["g"] = 218, ["b"] = 218, ["o"] = 255}
						end

						addButtonText(text, xpos, ypos, textScale, font)
						natives.GRAPHICS.DRAW_RECT(xpos, ypos, xscale, yscale, colors.r, colors.g, colors.b, colors.o);
					end
				else
					print("No buttons associated with menu: ", GUI.activeMenu)
					GUI.setActiveMenu("")
				end
			else
				print("No menu created with the name: ", GUI.activeMenu)
				GUI.setActiveMenu("")
			end
		end
	end

	function addButtonText(text, xpos, ypos, textScale, font)
		if(text ~= nil and xpos ~= nil and ypos ~= nil and textScale ~= nil and font ~= nil) then
			natives.UI.SET_TEXT_FONT(font)
			natives.UI.SET_TEXT_SCALE(0.0, textScale)
			natives.UI.SET_TEXT_COLOUR(255, 255, 255, 255)
			natives.UI.SET_TEXT_CENTRE(true)
			natives.UI.SET_TEXT_DROPSHADOW(0, 0, 0, 0, 0)
			natives.UI.SET_TEXT_EDGE(0, 0, 0, 0, 0)
			natives.UI._SET_TEXT_ENTRY("STRING")
			natives.UI._ADD_TEXT_COMPONENT_STRING(text)
			natives.UI._DRAW_TEXT(xpos, ypos - 0.0125)
		else
			print("text: ", text, ", xpos: ", xpos, ", ypos: ", ypos, ", textScale: ", textScale, ", font: ", font)
		    print("There seems to be a nil value when trying to add text to button (see above)")
		    GUI.setActiveMenu("")
		end
	end

	--this needs to be called every frame
	function GUI.tick()
		--this is pretty much a work around because using a while loop causes the game to freeze/crash
		--and using main:Wait(0) causes the text to flash instead of constant
		--this is called with the main:Run() loop
		GUI.drawStatusText()
		GUI.drawGUI()

		for i, key in pairs(GUI.navButtons.menuOpen.keys) do
			if(IsKeyDown(key)) then
				natives.AUDIO.PLAY_SOUND_FRONTEND(-1, "NAV_UP_DOWN", "HUD_FRONTEND_DEFAULT_SOUNDSET", false)
				if(GUI.activeMenu == GUI.navButtons.menuOpen.menu) then

					--if menu that is trying to be drawn is already drawn, go to the previous menu that is set for that menu 
					GUI.setActiveMenu(GUI.menu[GUI.activeMenu].prevMenu)
				else
					--if it isn't already drawn, draw as normal
				    GUI.setActiveMenu(GUI.navButtons.menuOpen.menu)
				    GUI.currentSelection = 1
				end
			end
		end

		--if the navUp key is pressed, move up (down numerically) in the menu
		for k, key in pairs(GUI.navButtons.navUp.keys) do
			if(IsKeyDown(key) and GUI.activeMenu ~= "") then
				GUI.activeButton = GUI.activeButton - 1
				natives.AUDIO.PLAY_SOUND_FRONTEND(-1, "NAV_UP_DOWN", "HUD_FRONTEND_DEFAULT_SOUNDSET", false)
				--loop selection to the last button if <= 0
				if(GUI.activeButton <= 0) then
					GUI.activeButton = GUI.menu[GUI.activeMenu].numButtons
				end
			end
		end

		--if the navDown key is pressed, move down (up numerically) in the menu
		for k, key in pairs(GUI.navButtons.navDown.keys) do
			if(IsKeyDown(key) and GUI.activeMenu ~= "") then
				GUI.activeButton = GUI.activeButton + 1
				natives.AUDIO.PLAY_SOUND_FRONTEND(-1, "NAV_UP_DOWN", "HUD_FRONTEND_DEFAULT_SOUNDSET", false)

				--loop selection to the first button if > the last button
				if(GUI.activeButton > GUI.menu[GUI.activeMenu].numButtons) then
					GUI.activeButton = 1
				end
			end
		end

		--if the select key is pressed, call the function the button was pressed on
		for k, key in pairs(GUI.navButtons.select.keys) do
			if(IsKeyDown(key) and GUI.activeMenu ~= "") then
				local currButton = GUI.buttons[GUI.activeMenu].settings[GUI.activeButton]
				natives.AUDIO.PLAY_SOUND_FRONTEND(-1, "NAV_UP_DOWN", "HUD_FRONTEND_DEFAULT_SOUNDSET", false)

				--if for some reason your function has more than 20 arguments, you can add more here
				if(type(currButton.funct) == "function" and type(currButton.args) == "table") then
					GUI.toggleActiveButtonState()
					currButton.funct(currButton.args[1], currButton.args[2], currButton.args[3], currButton.args[4], currButton.args[5], currButton.args[6], currButton.args[7], currButton.args[8], currButton.args[9], currButton.args[10], currButton.args[11], currButton.args[12], currButton.args[13], currButton.args[14], currButton.args[15], currButton.args[16], currButton.args[17], currButton.args[18], currButton.args[19], currButton.args[20])

				elseif(type(currButton.funct) == "function" and type(currButton.args) == "string" or type(currButton.args) == "number"or type(currButton.args) == "boolean") then
					GUI.toggleActiveButtonState()
					currButton.funct(currButton.args)
					
				elseif(currButton.args == nil) then
					GUI.toggleActiveButtonState()
					currButton.funct()

				else
					print("function: ", currButton.funct, ", args: ", currButton.args)
					print("Error: button \"", GUI.buttons[GUI.activeMenu].settings[GUI.activeButton].text, "\" doesn't seem to have a function or the arguments are invalid. Are you calling it correctly?" )
				end
			end
		end

		--if the navBack key is pressed, draw the prevMenu
		for k, key in pairs(GUI.navButtons.navBack.keys) do
			if(IsKeyDown(key) and GUI.activeMenu ~= "") then
				GUI.setActiveMenu(GUI.menu[GUI.activeMenu].prevMenu)
				natives.AUDIO.PLAY_SOUND_FRONTEND(-1, "NAV_UP_DOWN", "HUD_FRONTEND_DEFAULT_SOUNDSET", false);
			end
		end
	end

	return GUI
end

function rainSpawnTick()
	local models = {"WEAPON_VEHICLE_ROCKET", "WEAPON_FIREWORK", "MOLOTOV"}

	local randModel = models[math.random(#models)]
	local hash = natives.GAMEPLAY.GET_HASH_KEY(randModel)
	local plyCoords = natives.ENTITY.GET_ENTITY_COORDS(plyPed, true)
	local rand = {}
	rand.x = math.random(-100, 100)
	rand.y = math.random(-100, 100)
	rand.z = math.random(15, 35)

	local spawn = natives.ENTITY.GET_OFFSET_FROM_ENTITY_IN_WORLD_COORDS(plyPed, rand.x, rand.y, rand.z)

	natives.GAMEPLAY.SHOOT_SINGLE_BULLET_BETWEEN_COORDS(spawn.x, spawn.y, spawn.z, spawn.x, spawn.y, spawn.z - 100, 250, true, hash, plyPed, true, false, 10000.0);
end

function rainSpawnToggle()
	if(GUI.isActiveButtonToggledOn()) then
		rainSpawn = true
		GUI.updateStatusText("The sky begins to open...", 1000)
	else
		rainSpawn = false
		GUI.updateStatusText("The sky seems to clear...", 1000)
	end
end

function changeSpawnRate(num)
	if(rainSpawnRate + num < 45) then
		GUI.updateStatusText("Maximum spawn rate acheived (45ms)", 1000)
	else
		rainSpawnRate = rainSpawnRate + num
		GUI.updateStatusText("Spawn rate: " .. tostring(rainSpawnRate) .. "(ms)", 1000)
	end
end

function noclipToggle()
	if(GUI.isButtonToggledOn("player", 1)) then
		noclip = true
		natives.PED.SET_PED_CAN_RAGDOLL(plyPed, false)
		if(GUI.isButtonToggledOn("player", 2) == false) then
			noFallToggle()
			GUI.toggleButtonState("player", 2)
		end

		if(GUI.isButtonToggledOn("player", 3) == false) then
			godModeToggle()
			GUI.toggleButtonState("player", 3)
		end

		GUI.updateStatusText("Noclip, Godmode, and No Fall turned on!", 1000)
	else
	    noclip = false
		natives.PED.SET_PED_CAN_RAGDOLL(plyPed, true)

		if(GUI.isButtonToggledOn("player", 2) == true) then
			noFallToggle()
			GUI.toggleButtonState("player", 2)
		end

		if(GUI.isButtonToggledOn("player", 3) == true) then
			godModeToggle()
			GUI.toggleButtonState("player", 3)
		end

		GUI.updateStatusText("Noclip, Godmode, and No Fall turned off!", 1000)
	end
end

function gameNoclipListen()
	local veh = natives.PED.GET_VEHICLE_PED_IS_USING(plyPed)

	if(natives.PED.IS_PED_IN_VEHICLE(plyPed, veh, false) == false) then
		local speed = .3
		--SET_ENTITY_ROTATION(Entity entity, float Pitch, float Roll, float Yaw, Any p4, Any p5) // 0x0A345EFE
		local forward = natives.ENTITY.GET_ENTITY_FORWARD_VECTOR(plyPed)
		local coords = natives.ENTITY.GET_ENTITY_COORDS(plyPed, true)
		local roll = natives.ENTITY.GET_ENTITY_ROTATION(plyPed, 0)
		
		speedx = forward.x / 2
		speedy = forward.y / 2
		speedz = .5
		
		if(natives.CONTROLS.IS_CONTROL_PRESSED(2, 61)) then
			speedx = forward.x * 2
			speedy = forward.y * 2
			speedz = speedz * 2

		end

		--space
		if(natives.CONTROLS.IS_CONTROL_PRESSED(2, 22)) then
			natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x, coords.y, coords.z + speedz, false, false, true)
		end

		if(natives.ENTITY.GET_ENTITY_HEIGHT_ABOVE_GROUND(plyPed) > 1.5) then
			natives.ENTITY.STOP_ENTITY_ANIM(plyPed, 0, 0, 0)

			--ctrl
			if(natives.CONTROLS.IS_CONTROL_PRESSED(2, 36)) then
				natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x, coords.y, coords.z - speedz, false, false, true)
			end

			--w+shift
			if(natives.CONTROLS.IS_CONTROL_PRESSED(2, 32) and natives.CONTROLS.IS_CONTROL_PRESSED(2, 22)) then
				natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x + speedx, coords.y + speedy, coords.z + speedz, false, false, true)
			--w+ctrl
			elseif(natives.CONTROLS.IS_CONTROL_PRESSED(2, 32) and natives.CONTROLS.IS_CONTROL_PRESSED(2, 36)) then
				natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x + speedx, coords.y + speedy, coords.z - speedz, false, false, true)
			--w
			elseif(natives.CONTROLS.IS_CONTROL_PRESSED(2, 32)) then
				natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x + speedx, coords.y + speedy, coords.z, false, false, true)
			end
			
			--a
			if(natives.CONTROLS.IS_CONTROL_PRESSED(2, 34)) then
				natives.ENTITY.SET_ENTITY_ROTATION(plyPed, 0, 0, roll.z + 2, 0, false)
			end
			
			--s+shift
			if(natives.CONTROLS.IS_CONTROL_PRESSED(2, 31) and natives.CONTROLS.IS_CONTROL_PRESSED(2, 22)) then
				natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x - speedx, coords.y - speedy, coords.z + speedz, false, false, true)
			--s+ctrl
			elseif(natives.CONTROLS.IS_CONTROL_PRESSED(2, 31) and natives.CONTROLS.IS_CONTROL_PRESSED(2, 36)) then
				natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x - speedx, coords.y - speedy, coords.z - speedz, false, false, true)
			--s
			elseif(natives.CONTROLS.IS_CONTROL_PRESSED(2, 31)) then
				natives.ENTITY.SET_ENTITY_COORDS_NO_OFFSET(plyPed, coords.x - speedx, coords.y - speedy, coords.z, false, false, true)
			end

			--d
			if(natives.CONTROLS.IS_CONTROL_PRESSED(2, 35)) then
				natives.ENTITY.SET_ENTITY_ROTATION(plyPed, 0, 0, roll.z - 2, 0, false)
			end
		end
	end
end

function noFallToggle()
	if(noFall == false) then
		noFall = true
		GUI.updateStatusText("No Fall turned on!", 1000)
	else
		noFall = false
		natives.ENTITY.FREEZE_ENTITY_POSITION(plyPed, false)
		GUI.updateStatusText("No Fall turned off!", 1000)
	end
end

function gameNoFallListen()
	if(natives.PED.IS_PED_FALLING(plyPed) == true and noFall) then
		natives.ENTITY.FREEZE_ENTITY_POSITION(plyPed, true)
	else
		natives.ENTITY.FREEZE_ENTITY_POSITION(plyPed, false)
	end
end

function godModeToggle()
	if(godMode == false) then
		godMode = true
		natives.ENTITY.SET_ENTITY_MAX_HEALTH(plyPed, 0)
		natives.PLAYER.SET_PLAYER_INVINCIBLE(ply, true)
		GUI.updateStatusText("God Mode turned on!", 1000)
	else
		godMode = false
		natives.PLAYER.SET_PLAYER_INVINCIBLE(ply, false)
		GUI.updateStatusText("God Mode turned off!", 1000)
	end
end

function noWantedToggle()
	if(noWanted == false) then
		noWanted = true
	else
		noWanted = false
	end
end

--ran one time when the program starts up
function mainInit()
	--this only needs to be called once
	GUI = GUIinit()

	ply = natives.PLAYER.PLAYER_ID();
	plyPed = natives.PLAYER.PLAYER_PED_ID();

	noclip = false
	noFall = false
	godMode = false
	noWanted = false
	rainSpawn = false
	rainSpawnRate = 100

	GUI.createMenu("main", "", "Kmc's Mod Menu", 0.09, 0.1, 0.125, 0.0625, 0.005, 0.35, 4)
	GUI.addButton("main", "Player", GUI.setActiveMenu, "player", false, 0.35, 0)

	GUI.createMenu("player", "main", "Player", 0.09, 0.1, 0.125, 0.0625, 0.005, 0.35, 4)
	GUI.addButton("player", "Noclip", noclipToggle, nil, true, 0.35, 0)
	GUI.addButton("player", "No Fall", noFallToggle, nil, true, 0.35, 0)
	GUI.addButton("player", "God Mode", godModeToggle, nil, true, 0.35, 0)
	GUI.addButton("player", "No Wanted", noWantedToggle, nil, true, 0.35, 0)
	GUI.addButton("player", "Rain!", GUI.setActiveMenu, "rain", false, 0.35, 0)

	GUI.createMenu("rain", "player", "Super Top Secret", 0.09, 0.1, 0.125, 0.0625, 0.005, 0.35, 4)
	GUI.addButton("rain", "Make it rain!", rainSpawnToggle, nil, true, 0.35, 0)
	GUI.addButton("rain", "Decrease Rate", changeSpawnRate, 5, false, 0.35, 0)
	GUI.addButton("rain", "Increase Rate", changeSpawnRate, -5, false, 0.35, 0)

end

-- Run function
-- This is called once! It's your job to keep this alive!
function main:Run()
	-- ran once
	--mainInit is where set the buttons to be created (just to keep inside here somewhat uncluttered), look inside there for what you need to call to set up the GUI
	mainInit()
	--list of key codes https://msdn.microsoft.com/en-us/library/windows/desktop/dd375731%28v=vs.85%29.aspx
	local f5 = 0x74
	local numpad2 = 0x62
	local numpad8 = 0x68
	local numpad5 = 0x65
	local numpad0 = 0x60
	local closeBracket = 0xDD
	--menuOpen might be a litte confusing, but "menu" is what opens/closes when you press a button listed in "keys"
	--for example, I could have set it to "teleports", and the teleports menu would open
	--multiple keys look like this: {f5, numpad8}
	GUI.navButtons = {
						["menuOpen"] = {["keys"] = {f5}, ["menu"] = "main"},
						["navUp"] 	 = {["keys"] = {numpad8}}, 
						["navDown"]  = {["keys"] = {numpad2}}, 
						["select"] 	 = {["keys"] = {numpad5}}, 
						["navBack"]  = {["keys"] = {numpad0}}
					}

	-- ran until closed or crash
	while self:IsRunning() do
		--GUI.tick needs to be called every frame. if not, the menus won't appear
		GUI.tick()

		if(noclip == true) then
			gameNoclipListen()
		end

		if(noFall == true) then
			gameNoFallListen()
		end

		if(noWanted == true) then
			natives.PLAYER.CLEAR_PLAYER_WANTED_LEVEL(ply);
		end

		if(rainSpawn == true) then
			--this is basically a delay. by using this instead of using main:Wait(), it won't cause the whole script to pause
			if(nextSpawnTime == nil) then
				nextSpawnTime = natives.GAMEPLAY.GET_GAME_TIMER() + rainSpawnRate
			end

			if(nextSpawnTime < natives.GAMEPLAY.GET_GAME_TIMER()) then
				nextSpawnTime = nil
				rainSpawnTick()
			end
		end

		if(IsKeyDown(closeBracket)) then
			GUI.toggleButtonState("player", 1)
			noclipToggle()
		end
		
		self:Wait(0)
	end
end
function main:OnError() end
main:Register()